if (is3DEN) then {
	ttt_wireNumber = 0;
	add3DENEventHandler ["OnConnectingEnd", {
		params ["_class", "_from", "_to"];
		// systemChat format ["%1: %2 to %3", _class, _from, _to];
		if (isNil "_to") exitWith { systemChat "anchor connection aborted"; };
		private _fr = _from # 0 # 0;

		private _ret = false;
		// systemChat format ["ttt_wireNumber is %1", missionNamespace getVariable "ttt_wireNumber"];
		while {not _ret} do {
			private _primaryName = format ["ttt_wire_primary_%1", ttt_wireNumber];
			_fr set3DENAttribute ["name", _primaryName];
			if ((_fr get3DENAttribute "name") # 0 isNotEqualTo _primaryName) then { 
				systemChat "Wires: Setting name failed"; 
				// systemChat _primaryName;
				// systemChat str (_fr get3DENAttribute "name");
				_ret = false;
			} else {
				_ret = true;
			};

			private _secondaryName = format ["ttt_wire_secondary_%1", ttt_wireNumber];
			_to set3DENAttribute ["name", _secondaryName];
			if ((_to get3DENAttribute "name") # 0 isNotEqualTo _secondaryName) then { 
				systemChat "Wires: Setting name failed"; 
				_ret = false;
			} else {
				_ret = true;
			};
			ttt_wireNumber = ttt_wireNumber + 1;
			if (ttt_wireNumber >= 100) exitWith {
				"Did not find free wire number between 0 and 100" call bis_fnc_error;
			};
			// systemChat format ["_ret is %1", _ret];
		};
		if (_ret) then {
			systemChat format ["Wires: will create wires from %1 to %2 after mission start", _fr, _to];
		} else {
			systemChat "Wire setup aborted";
		};
	}];

} else {

	ttt_wires_fnc_setupWires = {
		params ["_anchor", "_syncedAnchor", "_wires_globals", "_key"];
		// Define a polygon that the wire may grapple within
		private _tolerance = 0.2;
		private _dir = (_anchor getDir _syncedAnchor) + 90; // roughly the runway direction
	    private _p1 = _anchor getPos [_tolerance, _dir];
	    _p1 set [2, getPosASL _anchor # 2];
	    private _p2 = _anchor getPos [-1 * _tolerance, _dir];
	    _p2 set [2, getPosASL _anchor # 2];
	    private _p3 = _syncedAnchor getPos [-1 * _tolerance, _dir];
	    _p3 set [2, getPosASL _syncedAnchor # 2];
	    private _p4 = _syncedAnchor getPos [_tolerance, _dir];
	    _p4 set [2, getPosASL _syncedAnchor # 2];

	    // Setup a center object that is attached to the aircraft hook later on
	    private _pos = _anchor getPos [(_anchor distance _syncedAnchor) / 2, _anchor getDir _syncedAnchor];
	    // _pos = _pos getPos [-0.5, getDir _carrier - 180];
	    _pos set [2, (getPosASL _anchor) # 2];
	    private _wireCenter = createVehicle ["ttt_wires_center", _pos, [], 0, "CAN_COLLIDE"];
	    _wireCenter setPosASL _pos;
	    private _ropeLeft = ropeCreate [_wireCenter, [0,0,0.1], _anchor, [0,0,0.05], (_wireCenter distance _anchor) - 0.1, 
	                            ["", [0,0,-1]], ["", [0,0,-1]], "ttt_wires_segment"];
	    private _ropeRight = ropeCreate [_wireCenter, [0,0,0.1], _syncedAnchor, [0,0,0.05], (_wireCenter distance _syncedAnchor) - 0.1, 
	                            ["", [0,0,-1]], ["", [0,0,-1]], "ttt_wires_segment"];

	    (_wires_globals get "ttt_wireAreas") set [_key, [_p1, _p2, _p3, _p4] ];
	    (_wires_globals get "ttt_wireObjects") set [_key, [_anchor, _syncedAnchor] ];
	    (_wires_globals get "ttt_wireRopes") set [_key, [_ropeLeft, _ropeRight] ];
	    (_wires_globals get "ttt_wireCenters") set [_key, _wireCenter];

	    systemChat "Setup wire complete";
	};

	ttt_wires_fnc_initLocal = {
		params ["_anchor"];

		if (isNil "ttt_wires_globals") then {
			ttt_wires_globals = createHashMap;
			ttt_wires_globals set ["ttt_wireAreas", createHashMap];
			ttt_wires_globals set ["ttt_wireObjects", createHashMap];
			ttt_wires_globals set ["ttt_wireRopes", createHashMap];
			ttt_wires_globals set ["ttt_wireCenters", createHashMap];
		};

		private _name = vehicleVarName _anchor;
		if ((_name find "secondary") != -1) exitWith { systemChat "Ignoring secondary anchor"; };
		// systemChat "1";
		private _number = (_name splitString "_") # 3;
		private _secondaryName = format ["ttt_wire_secondary_%1", _number];
		format ["Searching %1", _secondaryName] call bis_fnc_log;
		private _secondary = missionNamespace getVariable [_secondaryName, objNull];
		// systemChat "2";
		if (isNull _secondary) exitWith { "No secondary anchor assigned" call bis_fnc_error; };
		// systemChat "3";
		[_anchor, _secondary, ttt_wires_globals, format ["ttt_wires_%1", _number]] call ttt_wires_fnc_setupWires;
	};

	[] spawn {
	  	private _anchors = player nearObjects ["ttt_wires_anchor", 500];
		sleep 1;
		{
			systemChat format ["Calling init for %1", _x];
			_x spawn ttt_wires_fnc_initLocal;
		} forEach _anchors;
		sleep 1;
		"Begin wires setup log" call bis_fnc_log;
		{
			(ttt_wires_globals get _x) call bis_fnc_log;
		} forEach (keys ttt_wires_globals);
		systemChat "Logging complete";
	};


	ttt_wires_planes = [p];

	// [missionNamespace, "ttt_wires_toggleTailhook", ttt_wires_fnc_toggleTailhook] call BIS_fnc_addScriptedEventHandler;
	ttt_wires_fnc_hasTailhookDown = {
		/**
		 * Author:
		 * - TeTeT
		 * 
		 * Description:
		 * - Check if tailhook is in down position.
		 *   Two cases covered: simple tailhook down > 0 animationPhase for 'normal' planes
		 *                      Freedom compatible planes will have a more thorough check via config

		 * 
		 * Parameter(s):
		 *   _plane - the plane that will be checked for tailhook down
		 * 
		 * Returns:
		 *   true - the planes tailhook is down
		 *   false - the planes tailhook is up
		 **/
		params ["_plane"];
		// systemChat format ["DEBUG: plane %1 has tailhook %2", _plane, _plane animationPhase "tailhook"];
		// There was a reason to not use animationSourcePhase here, but I don't recall why...
		private _ret = (_plane animationPhase "tailhook") > 0;

		private _configPath = configFile >> "CfgVehicles" >> typeOf _plane;
		private _compat = _configPath >> "CarrierOpsCompatability";
		if (!(isNil "_compat") and !(isNull _compat)) then {
		    /**
		        F-181:
		        ArrestHookAnimationList[] = {"tailhook", "tailhook_door_l", "tailhook_door_r"};
		        ArrestHookAnimationStates[] = {0, 0.53, 1};
		    **/
		    private _planeHookAnimStates = (_configPath >> "CarrierOpsCompatability" >> "ArrestHookAnimationStates") call BIS_fnc_getCfgData;
		    _planeHookAnimStates params ["_planeHookDownAnimState","_planeHookCaughtAnimState","_planeHookUpAnimState"];
		    private _planeHookAnimList = (_configPath >> "CarrierOpsCompatability" >> "ArrestHookAnimationList") call BIS_fnc_getCfgData;
		    private _tailhookAnimation = _planeHookAnimList # 0;
		    _ret = ( (_plane animationPhase _tailhookAnimation) >= _planeHookCaughtAnimState );
		    if (_planeHookUpAnimState > 0) then {
		        _ret = ( (_plane animationPhase _tailhookAnimation) <= _planeHookCaughtAnimState );
		    };
		};
		_ret;
	};

	ttt_wires_fnc_touchdown = {
		/**
		 * Author:
		 * - TeTeT
		 * 
		 * Description:
		 * - Touchdown function when a plane lands on a carrier
		 *   Uses spawn as the braking method uses sleep, as well as the grappling method
		 *   TODO: Consider rewriting this as per frame eh on the client
		 * 
		 * Parameter(s):
		 *   _carrier - the carrier where the touchdown happened
		 *   _plane - the plane that landed
		 * 
		 * Returns:
		 *   true - always
		 **/
		 _this spawn {

			params ["_carrier", "_plane"];
			private _speed = speed _plane;

			if (player in _plane) then {addCamShake  [3.5, 2.8, 4.5];};
			_plane vehicleChat format ["Wire caught, with %1 km/h", _speed];

			// Depending on landing speed we alter the minimum speed for the grappling
			// E.g. high speed the wire will pull you back a bit
			//      low speed the plane will simply stop
			private _minSpeed = -10;
			if (_speed < 50) then {
				_minSpeed = 0;
			};
			
			// Copy from BI function fn_aircraftTailhook.sqf
			// default values from FA-181
			private _arrestVelocityReduction = -12;
			private _arrestSlowDownStep = 0.7;
			private _tailhookTension = 0.3;
			
			private _configPath = configFile >> "CfgVehicles" >> typeOf _plane;
			private _planeCarrierOpsEnabled = (_configPath >> "tailHook") call BIS_fnc_getCfgData; 
			if (_planeCarrierOpsEnabled isNotEqualTo 0) then {
				_arrestSlowDownStep = (_configPath >> "CarrierOpsCompatability" >> "ArrestSlowDownStep") call BIS_fnc_getCfgData;
				_arrestVelocityReduction = (_configPath >> "CarrierOpsCompatability" >> "ArrestVelocityReduction") call BIS_fnc_getCfgData;
				//NI: confiq interrogated for the tailhook animation states
				_tailhookTensionCfg = (_configPath >> "CarrierOpsCompatability" >> "ArrestHookAnimationStates") call BIS_fnc_getCfgData;
				_tailhookTension = _tailhookTensionCfg select 1;
			};
			//NI: animate tailhook upon wire catch to simulate tension, if CarrierOps enabled it will use config array, otherwise will animate for 0.3
			_plane animate ["tailhook",_tailhookTension];

			// This is quite a hack. The Freedom BIS_fnc_aircraftTailhook calls the below code but it's not braking hard enough for the Constellation.
			// so we apply some extra braking here, but only with a fraction of force, e.g. low _brakeFactor
			
			private _brakeFactor = 0.3;
			// if (_speed < 150) then {
			// 	_brakeFactor = 0.5;
			// } else {
			// 	_brakeFactor = _speed * 0.007 - 0.55; //_speed * 0.007 - 0.55
			// };
			// // NI: Limit braking gain by speed to 1.2
			// //150kph:(1.05); 175kph:(1.225); 200kph:(1.4); 225kph:(1.575); 250kph:(1.75);
			// _brakeFactor = _brakeFactor max 1.0;  //NI: previously "_brakeFactor min 1.2"
			// systemChat format ["DEBUG: Adjusting ArrestVelocityReduction dynamically by: %1", _brakeFactor];
			_arrestVelocityReduction = _arrestVelocityReduction * _brakeFactor;
			// systemChat format ["DEBUG: %1 ArrestSlowDownStep - %2 ArrestVelocityReduction", _arrestSlowDownStep, _arrestVelocityReduction];

			while {alive _plane and {_plane call ttt_wires_fnc_hasTailhookDown and {speed _plane > _minSpeed}}} do {
				// reduced the braking and leave it to BIS_fnc_aircraftTailhook
				_velInitial = velocity _plane;
				_dirPlane = direction _plane;
				_plane setVelocity (_velInitial vectorAdd [sin _dirPlane * _arrestVelocityReduction, cos _dirPlane * _arrestVelocityReduction, 0]);

				sleep _arrestSlowDownStep;
				if (_arrestVelocityReduction >= -3) then {_arrestVelocityReduction = -3;} else {_arrestVelocityReduction = _arrestVelocityReduction + 0.1;};
				if (_arrestSlowDownStep <= 0.2) then {_arrestSlowDownStep = 0.03;} else {_arrestSlowDownStep = _arrestSlowDownStep - 0.1;};
			};

			// did not work from the dedi server, grappling locally
			while  {alive _plane and {_plane call ttt_wires_fnc_hasTailhookDown}} do {
				_plane setVelocityModelSpace [0,0,0];
				sleep 0.01;
			};
			_carrier;
		};
		true;
	};

	ttt_wires_fnc_wireRopesAttach = {
		/**
		 * Author:
		 * - TeTeT
		 * 
		 * Description:
		 * - Attach ropes aka wire to the planes tailhook position
		 *   TODO: Consider rewriting this as per frame eh on the client
		 * 
		 * Parameter(s):
		 *   _carrier - the carrier where the touchdown happened
		 *   _plane - the plane that landed
		 * 
		 * Returns:
		 *   true - always
		 **/

		_this spawn 
		{
			params ["_carrier", "_plane"];

			systemChat str _carrier;
			format ["Carrier is %1", _carrier] call bis_fnc_log;
			// attach wire to tailhook

			private _tailhookAnimation = "tailhook";
			private _configPath = configFile >> "CfgVehicles" >> typeOf _plane;
			private _compat = _configPath >> "CarrierOpsCompatability";
			if (!(isNil "_compat") and !(isNull _compat)) then {
			    /**
			        F-181:
			        ArrestHookAnimationList[] = {"tailhook", "tailhook_door_l", "tailhook_door_r"};
			        ArrestHookAnimationStates[] = {0, 0.53, 1};
			    **/
			    _tailhookAnimation = (_configPath >> "CarrierOpsCompatability" >> "ArrestHookMemoryPoint") call BIS_fnc_getCfgData;
			};

			systemChat format ["Tailhook : %1", _tailhookAnimation];
			private _tailhookPos = _plane modelToWorld (_plane selectionPosition _tailhookAnimation);	
			private _key = _plane getVariable "ttt_wires_wire";
			((ttt_wires_globals get "ttt_wireObjects") get _key) params ["_wireLeft", "_wireRight"];
			((ttt_wires_globals get "ttt_wireCenters") get _key) params ["_wireCenter"];
			private _wireCenterOrigPos = getPosASL _wireCenter;
			_wireCenter attachTo [_plane, [0,0,0], _tailhookAnimation, true];
			while {alive _plane and (_plane call ttt_wires_fnc_hasTailhookDown)} do {
				private _tailhookPos = _plane modelToWorld (_plane selectionPosition _tailhookAnimation);
				if (_tailhookPos distance _wireLeft > 150 or 
					_tailhookPos distance _wireRight > 150) exitWith {
					_plane vehicleChat "SNAP - wire broke";
					// playSound3D ["ttt_wires_carrier\sounds\trapBreak.ogg", _plane, false, getPosASL _plane, 1.4, 1, 1000];
					_plane setVelocityModelSpace [0,0,0];
					detach _wireCenter;
					_wireCenter setPosASL _wireCenterOrigPos;
					sleep 10;
				};
				sleep 0.01;
			};
			detach _wireCenter;
			_wireCenter setPosASL _wireCenterOrigPos;
			systemChat "Wire's released";
		};
		true;
	};

	ttt_wires_fnc_addToLandingStats = {
		/**
		 * Author:
		 * - TeTeT
		 * 
		 * Description:
		 * - Note the landing statistics for the plane, in carrier and plane
		 *   There is potentially a race between get and set the stats variable, 
		 *   but as a plane will rarely land on two carriers or land twice, we neglect it
		 * 
		 * Parameter(s):
		 *   _carrier - the carrier where the touchdown happened
		 *   _plane - the plane that landed
		 * 
		 * Returns:
		 *   _carrier
		 **/
		params ["_carrier", "_plane"];

		private _wireName = _plane getVariable "ttt_wires_wire";
		private _tdSpeed = round speed _plane;
		systemChat format ["Landing: caught wire with %1 km/h", _tdSpeed];

		private _pilot = driver _plane;
		private _stat = [
			getPlayerUID _pilot,
			isPlayer _pilot,
			name _pilot,
			gettext (configfile >> "CfgVehicles" >> typeOf _plane >> "displayName"),
			_wireName, 
		    _tdSpeed,
		    date,
		    [ _carrier ] call BIS_fnc_getName
		];
		private _landingStats = _carrier getVariable ["ttt_wires_landingStats", []];
		private _playerLandingStats = player getVariable ["ttt_wires_landingStats", []];
		_landingStats pushBack _stat;
		_playerLandingStats pushBack _stat;
		// Save updated variable and broadcast it to all clients
		systemChat "Persisting landing performance to ttt_wires_landingStats";
		_carrier setVariable ["ttt_wires_landingStats", _landingStats, true];
		player setVariable ["ttt_wires_landingStats", _playerLandingStats];
		_carrier;
	};

	[missionNamespace, "ttt_wires_touchdown", ttt_wires_fnc_touchdown] call BIS_fnc_addScriptedEventHandler;
	[missionNamespace, "ttt_wires_touchdown", ttt_wires_fnc_wireRopesAttach] call BIS_fnc_addScriptedEventHandler;
	[missionNamespace, "ttt_wires_touchdown", ttt_wires_fnc_addToLandingStats] call BIS_fnc_addScriptedEventHandler;


	ttt_wires_fnc_caughtAnyWire = {
		/**
		 * Author:
		 * - TeTeT
		 * 
		 * Description:
		 * - check if the tailhookPos is close to any wire defined in the carrier
		 *   if close to a wire, plane is grappled and the wire name is stored in plane for later consumption
		 * 
		 * Parameter(s):
		 *   _plane - a plane to be checked for grappling with the carrier's wire
		 *   _tailhookPos - the tailhook position of the plane
		 * 
		 * Returns:
		 *   true - the plane was grappled to a wire
		 *   false - the plane is free to move
		 **/
		params ["_plane", "_tailhookPos"];

		private _wireAreas = ttt_wires_globals get "ttt_wireAreas";
		private _wires = keys _wireAreas;
		private _ret = false;
		{
			private _grappled = _tailhookPos inPolygon (_wireAreas get _x);
			// systemChat format ["DEBUG: %1 %2", _x, _grappled];
			if (_grappled) then {
				_ret = true;
				_plane setVariable ["ttt_wires_wire", _x];
				systemChat format ["DEBUG: %1 %2", _x, _grappled];
			};
		} forEach _wires;
		_ret;
	};

	ttt_wires_fnc_localWires = {
		/**
		 * Author:
		 * - TeTeT
		 * 
		 * Description:
		 * - Function called on the client where the plane parameter is local
		 *   Relies on two global array variables, ttt_wires_planes and ttt_wires_globals.
		 * 
		 * Parameter(s):
		 *   _wire - a ttt_constellation ship
		 *   _plane - a plane to be checked for grappling with the wire
		 * 
		 * Returns:
		 *   nothing
		 **/
		params ["_wireCenter", "_plane"];
		// systemChat format ["DEBUG: %1: %2", _wire, _plane];
		// systemChat "DEBUG: checking wire caught";
		if (!local _plane) exitWith {
			"Wires: Plane not local" call BIS_fnc_error;
		};
		
		private _tailhookPos = _plane modelToWorldWorld (_plane selectionPosition "tailhook");
		// diag_log format ["_tailhookPos: %1", _tailhookPos];
		// 2 meter vertical diff - a lot
		// systemChat format ["DEBUG: wire - %1", _wire];
		// systemChat format ["DEBUG: v diff %1", (_tailhookPos # 2) - (getPosATL _wire # 2)];
		if (not (_plane getVariable ["ttt_wires_grappled", false]) and
			// { (_tailhookPos # 2) - (getPosATL _wire # 2) < 2 and 
			// special treatment for USS Freedom and other high carriers
			{ getPosASL _plane # 2 < 30 and 
			{ [_plane, _tailhookPos] call ttt_wires_fnc_caughtAnyWire
			} } ) then {
			systemChat "Constellation arrest ropes: wire caught!";
			if (speed _plane > 270) then {
				_plane vehicleChat "Wire SNAPPED!";
				[missionNamespace, "ttt_wires_abort", [_wireCenter, _plane], false] call BIS_fnc_callScriptedEventHandler;
			} else {
				_plane setVariable ["ttt_wires_grappled", true, true];
				systemChat "DEBUG: Calling ttt_wires_touchdown";
				[missionNamespace, "ttt_wires_touchdown", [_wireCenter, _plane], false] call BIS_fnc_callScriptedEventHandler;
			};
		} else {
			diag_log format ["Wires arrest ropes: Not grappled - %1, %2, %3",not (_plane getVariable ["ttt_wires_grappled", false]), getPosASL _plane # 2 < 20,  [_plane, _tailhookPos] call ttt_wires_fnc_caughtAnyWire];			
		};
	};

	ttt_wires_fnc_wireRopes = {
		/**
		 * Author:
		 * - TeTeT
		 * 
		 * Description:
		 * - Function called per frame to check if plane gets grappled or ungrappled with wires.
		 *   Relies on two global array variables, ttt_wires_planes and ttt_wires_globals.
		 * 
		 * Parameter(s):
		 *   none
		 * 
		 * Returns:
		 *   nothing
		 **/
		// systemChat format ["DEBUG: EachFrame %1", time];
		{
			private _wireCenter = _y;
			{
				private _plane = _x;
				// systemChat "DEBUG: Checking plane and carrier";
				private _tailhookDown = _plane call ttt_wires_fnc_hasTailhookDown;
				if (not(_plane getVariable ["ttt_wires_grappled", false]) and 
					{  _tailhookDown and 
					{ _plane distance _wireCenter < 150 } }) then {
					// systemChat "DEBUG: Calling local wires script";
					[_wireCenter, _plane] remoteExec ["ttt_wires_fnc_localWires", _plane];
				};
				if (not _tailhookDown and 
					{ _plane getVariable ["ttt_wires_grappled", false] or 
					  _plane getVariable ["ttt_wires_grappledStatic",false] }) then {
					["Ungrappling %1", _plane] call BIS_fnc_logFormat;
					_plane setVariable ["ttt_wires_grappled", false, true];
					_plane setVariable ["ttt_wires_grappledStatic", false, true];
				};
			} forEach ttt_wires_planes;
		} forEach (ttt_wires_globals get "ttt_wireCenters");
	};

	["ttt_wires_perFrameHandler", "onEachFrame", {[] call ttt_wires_fnc_wireRopes;}, []] call BIS_fnc_addStackedEventHandler;
	
};

